###
### Download "Scipy" at
### https://sourceforge.net/projects/scipy/?source=typ_redirect
###

import numpy as np
from math import *
from scipy.stats import rankdata


###################################################
###step 0-A 起点から全ノードへの最小交通費用を計算######
###################################################
###0606劉くんのダイクストラ法プログラミングを使います。

#ラベル付き隣接行列:networkを表現
m = [
    [0,2,1,4,0],   # A (0)
    [2,0,0,3,0],   # B (1)
    [1,0,0,2,4],   # C (2)
    [4,3,2,0,1],   # D (3)
    [0,0,4,1,0],   # E (4)
    ]
#初期設定
max_cost = float('inf')      #無限大
m_node = len(m)              #ノード数は隣接行列によって求める
unchecked = [False] * m_node #未確定ノードの集合
cost = [max_cost] * m_node   #起点から各ノードへの最小費用
f_prev = [None] * m_node     #最短経路を列挙するための配列

def search(m,ori,des):     #ネットワークｍにおいて起点oriから目的地desまでの最短経路を探す
    cost[ori] = 0          # 起点のノードの距離は０とする
    f_prev[ori] = ori      #起点前のノードは起点とする
    now = ori              #現在地を起点とする
    
    while True:
        min = max_cost     #変数minは現段階の最小費用を表す（初期は無限大）
        next = -1          #nextは起点から最小費用のあるノードを表す(初期－1)
        unchecked[now] = True
        for i in range(m_node):        #変数iはノード（0~4）
            if unchecked[i]: continue  #ノードが確定しいない場合ループが続く
            if m[now][i]:              #今のノードiと起点(now=ori)が接続しているかどうか
                tmp_cost = m[now][i] + cost[now] #一時的の費用を計算し、最小費用の更新
                if cost[i] > tmp_cost:
                    cost[i] = tmp_cost
                    f_prev[i] = now    #直前のノードに更新
            if min > cost[i]:          #現段階の最小費用と最小費用を持つノードを更新
                min = cost[i]
                next = i
        now = next                    #確定された最小費用持つノードが新らしい”起点”となる
        if next == -1: break          #ノード番号が－1なるまで(すべてノードが確認される)
        
    print_path(f_prev, cost)          #各接続しているノード間ルートの費用
    return [get_path(ori, des, f_prev), cost[des]]

#結果の出力
def print_path(f_prev, cost):
    for i in range(len(f_prev)):
        print("%d, prev = %d, cost = %d" %  (i, f_prev[i], cost[i]))

def get_path(ori, des, f_prev):
    path = []
    now = des
    path.append(now)
    while True:
        path.append(f_prev[now])      # f_prevには一つ前のノードが入ってる。
        if f_prev[now] == ori: break  #格納されたノードを逆にたどっていけば経路が求められる
        now = f_prev[now]
    path.reverse()
    return path


###ここからDialのアルゴリズム用
origin=input("Origin?")
destination=input("Destination?")
tf=input("Total flow between O-D?")  #起点-終点の交通量
theta=1

path0, cost0 = search(m, origin, 0)
path1, cost1 = search(m, origin, 1)
path2, cost2 = search(m, origin, 2)
path3, cost3 = search(m, origin, 3)
path4, cost4 = search(m, origin, 4)


print('cost0-0:',cost0)
print('cost0-1:',cost1)
print('cost0-2:',cost2)
print('cost0-3:',cost3)
print('cost0-4:',cost4)
min_cost = [cost0,cost1,cost2,cost3,cost4]
print (min_cost)

rank = rankdata(min_cost) #min_costを昇順に並べた時の順位のリストをつくる
print ("rank", rank)


###################################################
###step 0-B 全リンクの尤度を計算######################
###################################################

ll = np.zeros((m_node,m_node))  #リンク尤度の行列：初期値0


for i in range(m_node):
    for j in range(m_node):
        if m[i][j] != 0:        #c(i)<c(j)の場合
            if min_cost[i] < min_cost[j]:
                ll[i][j] = exp(theta*(int(min_cost[j]) - int(min_cost[i]) - int(m[i][j])))
            else:               #c(i)<c(j)の場合
                ll[i][j] = 0

print ("link likelihood matrix is")
print (ll)                       #リンク尤度の行列：結果

###################################################
###step 1 各リンクの遷移確率(リンク・ウェイト)を計算####
###################################################


lw = np.zeros((m_node,m_node))      #リンク遷移確率の行列：初期値0
sum_lw = np.zeros((1,m_node))       #(計算過程用)リンク遷移確率和の行列：初期値0



for r in range(len(rank)):          #起点から近いノードの昇順=rank 
    for s in range(len(rank)):
        if rank[s] == r + 1:        #rankに該当するノードsを検索
            for h in range(m_node):
                if m[h][s] != 0:    #該当ノードsへの流入ノード
                    sum_lw[0][s] += lw[h][s]
            for j in range(m_node):                       
                if s == origin:     #起点の場合
                    lw[s][j] = ll[s][j]
                else:               #起点以外の場合
                    lw[s][j] = ll[s][j] * sum_lw[0][s]               
                    

print ("link weight matrix is")
print (lw)                            #リンク遷移確率の行列：結果


####################################################
###step 2 リンクウェイトから、リンク交通量の配分を計算###
####################################################

def link_flow(m):
	lf=np.zeros((m_node,m_node))                #リンク交通量配分の行列：初期値0
	sum_lf = np.zeros((1,m_node))               #(計算過程用)リンク交通量配分和の行列：初期値0

	for r in reversed(range(len(rank))):        #起点から近いノードの降順
		for s in reversed(range(len(rank))):
			if rank[s] == r + 1:                #rankに該当するノードsを検索
				for h in range(m_node):
					if m[s][h] != 0:            #該当ノードsからの流出ノード
						sum_lf[0][s] += lf[s][h]
				for i in range(m_node):
					if sum_lw[0][s] == 0:       #リンク交通量式の分母=0を除外
						pass
					else:
						if s == destination:    #終点の場合
							sum_lf[0][s] = tf
							lf[i][s] = sum_lf[0][s] * lw[i][s] / sum_lw[0][s]
						else:                   #終点以外の場合
							lf[i][s] = sum_lf[0][s] * lw[i][s] / sum_lw[0][s]
	return lf                

lf = link_flow(m)
print ("link flow matrix is")                  #リンク交通量配分の行列：結果
print (lf)


###
###おわり！
###

iteration = 1
param = 10
x_n = [row[:] for row in lf]
temp = [row[:] for row in x_n]
criterion = 0.05
convergence = 1
m_n = [row[:] for row in m]

fo=[
[0,(lf[0][1]/2000)**5,(lf[0][2]/3000)**5,(lf[0][3]/3000)**5,0],   # A (0)
[(lf[1][0]/2000)**5,0,0,(lf[1][3]/2000)**5,0],   # B (1)
[(lf[2][0]/3000)**5,0,0,(lf[2][3]/2000)**5,(lf[2][4]/2000)**5],   # C (2)
[(lf[3][0]/3000)**5,(lf[3][1]/2000)**5,(lf[3][2]/2000)**5,0,(lf[3][4]/3000)**5],   # D (3)
[0,0,(lf[4][2]/2000)**5,(lf[4][3]/3000)**5,0],   # E (4)
]


#CHANGE STOP CONDITION

#Preliminary: Copy m

while convergence>criterion and iteration < 100:
#Step 1: Set m_n a function of flow

    for i in range(0, len(m_n)):
        for j in range(0,len(m_n[0])):
            m_n[i][j] = (m[i][j] + fo[i][j])


#Step 2: Perform stochastic network flow algorithm on m_n for link flow matrix y_n
    y_n = link_flow(m_n)
    
    fo=[
    [0,(y_n[0][1]/2000)**5,(y_n[0][2]/3000)**5,(y_n[0][3]/3000)**5,0],   # A (0)
    [(y_n[1][0]/2000)**5,0,0,(y_n[1][3]/2000)**5,0],   # B (1)
    [(y_n[2][0]/3000)**5,0,0,(y_n[2][3]/2000)**5,(y_n[2][4]/2000)**5],   # C (2)
    [(y_n[3][0]/3000)**5,(y_n[3][1]/2000)**5,(y_n[3][2]/2000)**5,0,(y_n[3][4]/3000)**5],   # D (3)
    [0,0,(y_n[4][2]/2000)**5,(y_n[4][3]/3000)**5,0],   # E (4)
    ]
    
#Step 3: Set x_n = x_n + (1/n) * (y_n - x_n)
    for i in range(0, len(x_n)):
        for j in range(0,len(x_n[0])):
            temp[i][j] = x_n[i][j]
    for i in range(0, len(m_n)):
        for j in range(0,len(m_n[0])):
            x_n[i][j] = x_n[i][j] + (1 / iteration) * (y_n[i][j] - x_n[i][j])
    iteration = iteration + 1


print('MSA result for iteration ' + str(iteration))
print(x_n)
